package ru.dbotthepony.kstarbound.lua

import com.sun.jna.Callback
import com.sun.jna.Library
import com.sun.jna.Native
import com.sun.jna.Pointer
import com.sun.jna.ptr.LongByReference
import org.apache.logging.log4j.LogManager
import org.lwjgl.system.MemoryUtil
import java.io.Closeable
import java.io.PrintWriter
import java.io.StringWriter
import java.lang.RuntimeException
import java.nio.ByteBuffer
import java.nio.ByteOrder

private typealias size_t = Long
private typealias psize_t = LongByReference

private typealias lua_KContext = Pointer

interface ILua : Library {
	fun luaL_newstate(): Pointer
	fun lua_checkstack(luaState: Pointer, amount: Int): Int
	fun lua_absindex(luaState: Pointer, value: Int): Int

	fun lua_close(luaState: Pointer)
	fun lua_atpanic(luaState: Pointer, fn: lua_CFunction): Pointer

	interface lua_CFunction : Callback {
		fun invoke(luaState: Pointer): Int
	}

	interface lua_Reader : Callback {
		fun invoke(luaState: Pointer, data: Pointer?, size: psize_t): Pointer? /* String */
	}

	fun lua_load(luaState: Pointer, reader: lua_Reader, data: Pointer?, chunkName: String, mode: String): Int

	// Оно является макросом
	/* fun lua_pcall(luaState: Pointer, numArgs: Int, numResults: Int, msgh: Int): Int {
		return lua_pcallk(luaState, numArgs, numResults, msgh, Pointer.NULL, null)
	} */

	interface lua_KFunction : Callback {
		/*
		** Type for continuation functions
		*/
		fun invoke(luaState: Pointer, status: Int, ctx: lua_KContext?): Int
	}

	fun lua_pcallk(luaState: Pointer, numArgs: Int, numResults: Int, msgh: Int, ctx: lua_KContext?, k: lua_KFunction?): Int
	fun lua_callk(luaState: Pointer, numArgs: Int, numResults: Int, ctx: lua_KContext?, k: lua_KFunction?): Int

	fun lua_tolstring(luaState: Pointer, index: Int, len: psize_t): Pointer?

	fun lua_pushcclosure(luaState: Pointer, fn: lua_CFunction, n: Int)

	fun lua_setglobal(luaState: Pointer, name: String)

	// Стандартные библиотеки
	fun luaopen_base(luaState: Pointer)
	fun luaopen_package(luaState: Pointer)
	fun luaopen_coroutine(luaState: Pointer)
	fun luaopen_table(luaState: Pointer)
	fun luaopen_io(luaState: Pointer)
	fun luaopen_os(luaState: Pointer)
	fun luaopen_string(luaState: Pointer)
	fun luaopen_math(luaState: Pointer)
	fun luaopen_utf8(luaState: Pointer)
	fun luaopen_debug(luaState: Pointer)

	fun lua_pushlstring(luaState: Pointer, str: Pointer, len: size_t): Pointer?
	fun lua_error(luaState: Pointer): Int
	fun lua_gettop(luaState: Pointer): Int
}

val LUA_JNA: ILua = Native.load("lua54", ILua::class.java)

private fun stringToBuffer(str: String): ByteBuffer {
	val bytes = str.toByteArray(charset = Charsets.UTF_8)
	val buf = ByteBuffer.allocateDirect(bytes.size)
	buf.order(ByteOrder.nativeOrder())
	bytes.forEach(buf::put)
	buf.position(0)

	return buf
}

private val SHARED_BUFFER = ByteBuffer.allocateDirect(16384).also { it.order(ByteOrder.nativeOrder()) }
private val SHARED_BUFFER_PTR = Pointer(MemoryUtil.memAddress(SHARED_BUFFER))

private fun loadString(string: String): Int {
	SHARED_BUFFER.position(0)
	val bytes = string.toByteArray(charset = Charsets.UTF_8)
	bytes.forEach(SHARED_BUFFER::put)
	SHARED_BUFFER.position(0)
	return bytes.size
}

private fun lua_pushlstring(luaState: Pointer, str: String) {
	val len = loadString(str)
	LUA_JNA.lua_pushlstring(luaState, SHARED_BUFFER_PTR, len.toLong())
}

class LuaState : Closeable {
	private val pointer = LUA_JNA.luaL_newstate()
	private val nativePointer get() = Pointer.nativeValue(pointer)

	private var destroyed = false
	private val panicHandler = object : ILua.lua_CFunction {
		override fun invoke(luaState: Pointer): Int {
			throw RuntimeException("$this panicked!")
		}
	}

	init {
		if (pointer == Pointer.NULL)
			throw OutOfMemoryError("Unable to allocate new LuaState")

		LUA_JNA.lua_atpanic(pointer, panicHandler)

		LUA_JNA.luaopen_base(pointer)
		LUA_JNA.luaopen_package(pointer)
		LUA_JNA.luaopen_table(pointer)
		LUA_JNA.luaopen_coroutine(pointer)
		LUA_JNA.luaopen_string(pointer)
		LUA_JNA.luaopen_math(pointer)
		LUA_JNA.luaopen_utf8(pointer)
		LUA_JNA.luaopen_debug(pointer)

		pushClosure {
			val build = mutableListOf<String?>()

			for (i in 1 .. stackTop) {
				build.add(getString(i))
			}

			LOGGER.info("Lua/print(): {}", build.joinToString("\t"))
		}

		LUA_JNA.lua_setglobal(pointer, "print")
	}

	fun pushClosure(lambda: (state: LuaState) -> Unit) {
		LuaJNI.lua_pushcclosure(nativePointer) lazy@{
			try {
				lambda.invoke(this@LuaState)
			} catch (err: Throwable) {
				val builder = StringWriter()
				val printWriter = PrintWriter(builder)
				err.printStackTrace(printWriter)
				lua_pushlstring(pointer, builder.toString())
				return@lazy 1
			}

			return@lazy 0
		}
	}

	fun checkStack(minAmount: Int): Boolean {
		return LUA_JNA.lua_checkstack(pointer, minAmount) > 0
	}

	fun absoluteIndex(index: Int): Int {
		return LUA_JNA.lua_absindex(pointer, index)
	}

	fun load(code: String, chunkName: String = "main chunk") {
		val buf = stringToBuffer(code)

		throwLoadError(LUA_JNA.lua_load(pointer, object : ILua.lua_Reader {
			override fun invoke(luaState: Pointer, data: Pointer?, size: psize_t): Pointer? {
				if (buf.remaining() == 0) {
					size.value = 0
					return Pointer.NULL
				}

				size.value = buf.remaining().toLong()
				val p = Pointer(MemoryUtil.memAddress(buf))
				buf.position(buf.remaining())
				return p
			}
		}, Pointer.NULL, chunkName, "t"))
	}

	fun pcall(numArgs: Int = 0, numResults: Int = 0): Int {
		val status = LuaJNA.lua_pcallk(nativePointer, numArgs, numResults, 0, null, null)

		if (status == LUA_ERRRUN) {
			throw LuaRuntimeException(getString())
		}

		return status
	}

	fun getString(index: Int = -1, limit: Long = 4096): String? {
		val len = psize_t()
		val p = LUA_JNA.lua_tolstring(pointer, absoluteIndex(index), len) ?: return null

		if (len.value == 0L) {
			return ""
		}

		if (len.value >= limit) {
			throw IllegalStateException("Unreasonably long Lua string: ${len.value}")
		}

		val readBytes = ByteArray(len.value.toInt())
		p.read(0L, readBytes, 0, readBytes.size)
		return readBytes.toString(charset = Charsets.UTF_8)
	}

	val stackTop get() = LUA_JNA.lua_gettop(pointer)

	override fun close() {
		if (destroyed) {
			throw IllegalStateException("Already destroyed")
		}

		LUA_JNA.lua_close(pointer)
		destroyed = true
	}

	companion object {
		private val LOGGER = LogManager.getLogger(LuaState::class.java)
	}
}
