package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.client.ClientWorld
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d
import ru.dbotthepony.kstarbound.util.Color
import ru.dbotthepony.kstarbound.world.World

enum class Move {
	STAND_STILL,
	MOVE_LEFT,
	MOVE_RIGHT
}

open class AliveEntity(world: World<*, *>) : Entity(world) {
	open var maxHealth = 10.0
	open var health = 10.0
	open val moveDirection = Move.STAND_STILL
	override val collisionResolution = CollisionResolution.SLIDE

	open val aabbDucked get() = aabb

	override val currentaabb: AABB get() {
		if (isDucked) {
			return aabbDucked
		}

		return super.currentaabb
	}

	var wantsToDuck = false
	var isDucked = false
		protected set

	/**
	 * Максимальная скорость передвижения этого существа в Starbound Units/секунда
	 *
	 * Скорость передвижения: Это скорость вдоль земли (или в воздухе, если парит) при ходьбе.
	 *
	 * Если вектор скорости вдоль поверхности (или в воздухе, если парит) больше заданного значения,
	 * то сущность быстро тормозит (учитывая силу трения)
	 */
	open val topSpeed = 20.0

	/**
	 * Скорость ускорения сущности в Starbound Units/секунда^2
	 *
	 * Если сущность хочет двигаться вправо или влево (а также вверх или вниз, если парит),
	 * то она разгоняется с данной скоростью.
	 */
	open val moveSpeed = 64.0

	/**
	 * То, как сущность может влиять на свою скорость в Starbound Units/секунда^2
	 * когда находится в свободном падении или в состоянии невесомости.
	 *
	 * Позволяет в т.ч. игрокам изменять свою траекторию полёта в стиле Quake.
	 */
	open val freeFallMoveSpeed = 8.0

	/**
	 * "Сила", с которой сущность останавливается, если не хочет двигаться.
	 *
	 * Зависит от текущего трения, так как технически является множителем трения поверхности,
	 * на которой стоит сущность. Если сущность парит, то сила трения является константой и не зависит от её окружения.
	 */
	open val brakeForce = 32.0

	/**
	 * Импульс прыжка данной сущности. Если сущность парит, то данное значение не несёт никакой
	 * полезной нагрузки.
	 */
	open val jumpForce = 20.0

	/**
	 * Высота шага данной сущности. Данное значение отвечает за то, на сколько блоков
	 * сможет подняться сущность просто двигаясь в одном направлении без необходимости прыгнуть.
	 */
	open val stepSize = 1.1

	override fun thinkPhysics(delta: Double) {
		super.thinkPhysics(delta)
		thinkMovement(delta)
	}

	protected var jumpRequested = false
	protected var nextJump = 0.0

	open fun requestJump() {
		if (jumpRequested || nextJump > world.timer)
			return

		jumpRequested = true
		nextJump = world.timer + 0.1
	}

	open fun recallJump() {
		if (jumpRequested) {
			jumpRequested = false
			nextJump = 0.0
		}
	}

	open fun thinkMovement(delta: Double) {
		if (onGround || !affectedByGravity) {
			var add: Vector2d

			if (isDucked) {
				thinkFriction(delta * brakeForce)
				add = Vector2d.ZERO
			} else {
				when (moveDirection) {
					Move.STAND_STILL -> {
						thinkFriction(delta * brakeForce)
						add = Vector2d.ZERO
					}

					Move.MOVE_LEFT -> {
						if (velocity.x > 0.0) {
							thinkFriction(delta * brakeForce)
						}

						add = Vector2d(x = -delta * moveSpeed)
					}

					Move.MOVE_RIGHT -> {
						if (velocity.x < 0.0) {
							thinkFriction(delta * brakeForce)
						}

						add = Vector2d(x = delta * moveSpeed)
					}
				}
			}

			if (add != Vector2d.ZERO) {
				if (isSpaceOpen(add, delta)) {
					velocity += add

					// спускание с "лестницы"
					val sweep = sweepAbsolute(pos + velocity * delta * 4.0, Vector2d(y = -stepSize), delta)

					if (sweep.hitAnything && sweep.hitPosition.y < -0.1) {
						dropToFloor()
						thinkFriction(delta * 6.0)
					}

					if (world is ClientWorld && world.client.settings.debugCollisions) {
						world.client.onPostDrawWorldOnce {
							world.client.gl.quadWireframe(worldaabb + velocity * delta * 4.0 + sweep.hitPosition, Color.RED)
						}
					}
				} else {
					// подъем по "лестнице"
					val sweep = sweepRelative(Vector2d(y = stepSize), delta)

					if (!sweep.hitAnything) {
						val sweep2 = sweepAbsolute(pos + sweep.hitPosition, Vector2d(x = -0.1 + add.x), delta)

						if (!sweep2.hitAnything) {
							pos += sweep.hitPosition + sweep2.hitPosition
							thinkFriction(delta * 64.0)
							dropToFloor()
						}

						if (world is ClientWorld && world.client.settings.debugCollisions) {
							world.client.onPostDrawWorldOnce {
								world.client.gl.quadWireframe(worldaabb + sweep.hitPosition + sweep2.hitPosition, Color.GREEN)
							}
						}
					}

					if (world is ClientWorld && world.client.settings.debugCollisions) {
						world.client.onPostDrawWorldOnce {
							world.client.gl.quadWireframe(worldaabb + sweep.hitPosition, Color.BLUE)
						}
					}
				}
			}

			if (velocity.length > topSpeed) {
				thinkFriction(delta * (velocity.length - topSpeed))
			}

			if (jumpRequested) {
				jumpRequested = false
				velocity += groundNormal * jumpForce
				onGround = false
			}
		} else if (!onGround && affectedByGravity) {
			when (moveDirection) {
				Move.STAND_STILL -> {}

				Move.MOVE_LEFT -> {
					val add = Vector2d(x = -delta * freeFallMoveSpeed)

					if (isSpaceOpen(add, delta))
						velocity += add
				}

				Move.MOVE_RIGHT -> {
					val add = Vector2d(x = delta * freeFallMoveSpeed)

					if (isSpaceOpen(add, delta))
						velocity += add
				}
			}
		}

		if (wantsToDuck && onGround) {
			isDucked = true
		} else if (isDucked) {
			if (world.isSpaceEmptyFromTiles(aabb + pos)) {
				isDucked = false
			}
		}
	}
}
