package ru.dbotthepony.kstarbound.world.entities

import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.client.render.EntityRenderer
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d
import ru.dbotthepony.kstarbound.math.lerp
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.World

enum class CollisionResolution {
	STOP,
	BOUNCE,
	PUSH,
	SLIDE,
}

/**
 * Определяет из себя сущность в мире, которая имеет позицию, скорость и коробку столкновений
 */
open class Entity(val world: World<*, *>) {
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (!isSpawned) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			}

			if (value == field) {
				return
			}

			val chunkPos = ChunkPos.fromTilePosition(pos)

			if (value != null && chunkPos != value.pos) {
				throw IllegalStateException("Set proper position before setting chunk this Entity belongs to")
			}

			val oldChunk = field
			field = value

			if (oldChunk == null && value != null) {
				world.orphanedEntities.remove(this)
				value.addEntity(this)
			} else if (oldChunk != null && value == null) {
				world.orphanedEntities.add(this)
				oldChunk.removeEntity(this)
			} else if (oldChunk != null && value != null) {
				value.transferEntity(this, oldChunk)
			}
		}

	open val currentaabb: AABB get() = aabb
	open val worldaabb: AABB get() = currentaabb + pos

	var pos = Vector2d()
		set(value) {
			if (field == value)
				return

			val old = field
			field = value

			if (isSpawned) {
				val oldChunkPos = ChunkPos.fromTilePosition(old)
				val newChunkPos = ChunkPos.fromTilePosition(value)

				if (oldChunkPos != newChunkPos) {
					chunk = world.getChunk(newChunkPos)?.chunk
				}
			}
		}

	var velocity = Vector2d()
	var isSpawned = false
		private set
	var isRemoved = false
		private set

	fun spawn() {
		if (isSpawned)
			throw IllegalStateException("Already spawned")

		isSpawned = true
		world.entities.add(this)
		chunk = world.getChunk(ChunkPos.fromTilePosition(pos))?.chunk

		if (chunk == null) {
			world.orphanedEntities.add(this)
		}
	}

	fun remove() {
		if (isRemoved)
			throw IllegalStateException("Already removed")

		isRemoved = true

		if (isSpawned) {
			world.entities.remove(this)
			chunk?.removeEntity(this)
		}
	}

	/**
	 * Касается ли сущность земли
	 *
	 * Данный флаг выставляется при обработке скорости, если данный флаг не будет выставлен
	 * правильно, то сущность будет иметь очень плохое движение в стороны
	 *
	 * Так же от него зависит то, может ли сущность двигаться, если она не парит
	 *
	 * Если сущность касается земли, то на неё не действует гравитация
	 */
	var onGround = false
		protected set(value) {
			field = value
			nextOnGroundUpdate = world.timer + 0.1
		}

	protected var nextOnGroundUpdate = 0.0

	var groundNormal = Vector2d.ZERO
		protected set

	protected var isHuggingAWall = false

	// наследуемые свойства
	open val aabb = AABB.rectangle(Vector2d.ZERO, 0.9, 0.9)
	open val affectedByGravity = true
	open val collisionResolution = CollisionResolution.STOP

	protected open fun onTouchGround(velocity: Vector2d, normal: Vector2d) {

	}

	protected fun sweepRelative(velocity: Vector2d, delta: Double, collisionResolution: CollisionResolution = this.collisionResolution) = world.sweep(worldaabb, velocity, collisionResolution, delta)
	protected fun sweepAbsolute(from: Vector2d, velocity: Vector2d, delta: Double, collisionResolution: CollisionResolution = this.collisionResolution) = world.sweep(aabb + from, velocity, collisionResolution, delta)
	protected fun isSpaceOpen(relative: Vector2d, delta: Double) = !sweepRelative(relative, delta).hitAnything

	fun dropToFloor() {
		val sweep = sweepRelative(Vector2d.DROP_TO_FLOOR, 1.0, CollisionResolution.STOP)

		if (!sweep.hitAnything)
			return

		pos += sweep.hitPosition
	}

	protected open fun propagateVelocity(delta: Double) {
		if (velocity.length == 0.0)
			return

		val sweep = sweepRelative(velocity * delta, delta)
		this.velocity = sweep.hitPosition / delta
		this.pos += this.velocity * delta

		if (nextOnGroundUpdate <= world.timer || !onGround) {
			onGround = sweep.hitNormal.dotProduct(world.gravity.normalized) <= -0.98
			groundNormal = sweep.hitNormal

			if (!onGround) {
				val sweepGround = sweepRelative(world.gravity * delta, delta)
				onGround = sweepGround.hitAnything && sweepGround.hitNormal.dotProduct(world.gravity.normalized) <= -0.98
				groundNormal = sweepGround.hitNormal
			}
		}
	}

	protected open fun thinkGravity(delta: Double) {
		velocity += world.gravity * delta
	}

	protected open fun thinkFriction(delta: Double) {
		velocity *= Vector2d(lerp(delta, 1.0, 0.01), 1.0)
	}

	protected open fun thinkPhysics(delta: Double) {
		if (!onGround && affectedByGravity)
			thinkGravity(delta)

		propagateVelocity(delta)

		if (affectedByGravity && onGround)
			thinkFriction(delta)

		//dropToFloor()
	}

	/**
	 * Заставляет сущность "думать".
	 */
	fun think(delta: Double) {
		if (!isSpawned) {
			throw IllegalStateException("Tried to think before spawning in world")
		}

		thinkPhysics(delta)
		thinkAI(delta)
	}

	protected open fun thinkAI(delta: Double) {

	}

	companion object {
		private val LOGGER = LogManager.getLogger(Entity::class.java)
	}
}
